/*
// \file extensions/wiicontroller/entityplugins/wiicontrolled.cpp
//
// Description:
// An entity plugin that maps the orientation and position of the connected
// Wii-controller to selected entities in the scene.
//
// \Author Anders Sabinsky Tgern <ast@hoast.dk>
// \Author Lasse Jon Fuglsang Pedersen <fuglsang@diku.dk>
//
// \Copyright See COPYING file that comes with the 3DOT distribution
*/

#include "wiicontrolled.h"
#include <3DOT/extensions/wiicontroller/iwiicontrollerextension.h>
#include <3DOT/handles/extensionhandle.h>
#include <entity/event.h>
#include <3DOT/extensions/base_interfaces/interfaces/itransformable.h>

namespace EXT_WIICONTROLLER_3DOT
{
	/*
	 * Setup
	 */

	// Constructor
	WiiControlled::WiiControlled(CORE_3DOT::HEntityPluginFactory factory)
		:
		CORE_3DOT::BaseEntityPlugin<IWiiControlled>(factory),
		mOrientationChangedHandler(new CORE_3DOT::BaseEventHandler(this,
			&WiiControlled::OnOrientationChanged)),
		mPositionChangedHandler(new CORE_3DOT::BaseEventHandler(this,
			&WiiControlled::OnPositionChanged)),
		mWiiControlledOrientation(false),
		mWiiControlledPosition(false)
	{
	}

	// This function is called when an instance of the
	// entity plugin is added to an entity
	void WiiControlled::OnAddedToEntity()
	{
		HWiiControllerExtension()->GetOrientationChangedEvent()
			->AddHandler(mOrientationChangedHandler);
		HWiiControllerExtension()->GetPositionChangedEvent()
			->AddHandler(mPositionChangedHandler);
	}

	// This function is called when serialization is run
	void WiiControlled::Serialize(CORE_3DOT::HSerializeStream s)
	{
		this->BaseEntityPlugin<IWiiControlled>::Serialize(s);
		// Her sker spndende ting:
		// - Hvis der loades sttes de angivne variable.
		// - Hvis der gemmes aflses de angivne variable.
		//s->attr("position", mPosition, true);
		//s->attr("orientation", mOrientation, true );
		//s->attr("scale", mScale, true);
	}

	// This function is called after serialization of ALL objects has run
	void WiiControlled::Update()
	{
		HWiiControllerExtension()->GetOrientationChangedEvent()
			->AddHandler(mOrientationChangedHandler);
		HWiiControllerExtension()->GetPositionChangedEvent()
			->AddHandler(mPositionChangedHandler);
	}

	/*
	 * Flags
	 */

	void WiiControlled::SetWiiControlledOrientation(bool flag)
	{
		mWiiControlledOrientation = flag;
	}

	bool WiiControlled::GetWiiControlledOrientation() const
	{
		return mWiiControlledOrientation;
	}

	void WiiControlled::SetWiiControlledPosition(bool flag)
	{
		mWiiControlledPosition = flag;
	}

	bool WiiControlled::GetWiiControlledPosition() const
	{
		return mWiiControlledPosition;
	}

	/*
	 * Event handlers
	 */

	// This function is called by WiiControllerExtension;
	// more specifically the orientation changed event
	void WiiControlled::OnOrientationChanged(CORE_3DOT::HRTTIUnknown sender,
											 CORE_3DOT::HRTTIUnknown data)
	{
		// If entity plugin is set to make use of reported orientation
		if (mWiiControlledOrientation)
		{
			// Get transformable entity
			EXT_BASEINTERFACES_3DOT::HTransformable t = GetEntity();
			
			// Get reported orientation
			CORE_3DOT::quat orientation = HWiiControllerExtension()
				->GetOrientation();

			// Map reported orientation
			t->SetOrientationLocal(orientation);
		}
	}

	// This function is called by WiiControllerExtension;
	// more specifically the position changed event
	void WiiControlled::OnPositionChanged(CORE_3DOT::HRTTIUnknown sender,
										  CORE_3DOT::HRTTIUnknown data)
	{
		// If entity plugin is set to make use of reported position
		if (mWiiControlledPosition)
		{
			// Get transformable entity
			EXT_BASEINTERFACES_3DOT::HTransformable t = GetEntity();
			
			// Get reported position
			CORE_3DOT::vec3 position = HWiiControllerExtension()->GetPosition();

			// Map reported position
			t->SetPositionLocal(position);
		}
	}
}
