/*
// \file extensions/wiicontroller/wiicontrollerextension.cpp
//
// Description:
// Connects to- and exposes data and events from a Wii-controller.
//
// \Author Anders Sabinsky Tgern <ast@hoast.dk>
// \Author Lasse Jon Fuglsang Pedersen <fuglsang@diku.dk>
//
// \Copyright See COPYING file that comes with the 3DOT distribution
*/

#define _USE_MATH_DEFINES 

#include "wiicontrollerextension.h"
#include "wiicontrollertask.h"
#include "entityplugins/wiicontrolled.h"
#include <entity/event.h>
#include <3DOT/extensions/base_interfaces/interfaces/itransformable.h>
#include <cmath>

#define DELTA_T	33 // 30 frames per second ~= 33ms between samples

namespace EXT_WIICONTROLLER_3DOT
{
	// Exported symbol (3DOT DLL entry point)
	DLLEXPORT_C CORE_3DOT::IExtension*
	getExtension(APPLICATION_3DOT::IApplication* app)
	{
		return new WiiControllerExtension(app);
	}

	/*
	 * Setup
	 */

	// Constructor, exports
	WiiControllerExtension::WiiControllerExtension(
		APPLICATION_3DOT::HApplication app)
		:
		CORE_3DOT::BaseExtension<IWiiControllerExtension>(app),
		mButtonStatesChangedEvent(new CORE_3DOT::BaseEvent()),
		mOrientationChangedEvent(new CORE_3DOT::BaseEvent()),
		mPositionChangedEvent(new CORE_3DOT::BaseEvent()),
		mEstTilt(WiiLib::Filter::CreateGaussFilter(1),
				 new WiiLib::GravityNaive()),
		mEstPosition(WiiLib::Filter::CreateGaussFilter(1),
					 new WiiLib::IntegratorMidpoint())
	{
		// Requirements
		AddRequiredEntityPlugin<EXT_BASEINTERFACES_3DOT::ITransformable>();

		// Exports
		AddExportedScheduleTaskWithFactory<IWiiControllerTask,
											WiiControllerTask>( );
		AddExportedEntityPluginWithFactory<IWiiControlled,
											WiiControlled>( );
	}

	// Destructor
	WiiControllerExtension::~WiiControllerExtension()
	{
		// Disconnect if connected
	    if (mController.IsConnected())
	    {
	        mController.Disconnect();
		}
	}

	// Initialize
	void WiiControllerExtension::Initialize(APPLICATION_3DOT::HApplication app,
											EDITOR_3DOT::HEditor editor)
	{
		// Attempt to connect on startup
		ConnectAndCalibrate();
	}

	/*
	 * Processing
	 */

	bool WiiControllerExtension::ConnectAndCalibrate()
	{
		if (!mController.IsConnected())
		{
			// Stop sending anything but button changes for now
			mController.SetReportMode(Controller::Report_Buttons, false);

			// Abort if unable to connect
			if (!mController.Connect())
			{
				return false;
			}

			LOG("WiiControllerExtension") << "Connected." << LOGEND();

			// Wait for calibration once connected
			while (!mController.IsCalibrated())
			{
				mController.Sample();
				mController.CalibrateAsync();
			}

			LOG("WiiControllerExtension") << "Calibrated." << LOGEND();

			// Start sending button changes and motion data continuously
			mController.SetReportMode(Controller::Report_ButtonsMotion, true);
			mController.SetLEDs(LED_1|LED_2|LED_3|LED_4);
		}

		// Connected and calibrated
		return true;
	}

	// Returns a CORE_3DOT::vec3 from a WiiLib::Vec3
	CORE_3DOT::vec3 WiiControllerExtension::ConvWiiLibVec3(WiiLib::Vec3 &v)
	{
		return CORE_3DOT::vec3(v.mX, v.mY, v.mZ);
	}

	// This function polls controller class for new data and fires all events
	void WiiControllerExtension::Process()
	{
		int				wiiLibCode;
		WiiLib::Vec3	wiiLibAxis;
		double			wiiLibAngle;
		WiiLib::Vec3	wiiLibPosition;
		WiiLib::Vec3	wiiLibGravity;

		CORE_3DOT::quat	newOrientation;
		CORE_3DOT::vec3	newPosition;

		// Attempt to connect if not connected
		if (!mController.IsConnected())
		{
			// Abort process if unable to connect
			if (!ConnectAndCalibrate())
			{
				return;
			}
		}

		// Instruct controller class to read next sample
		wiiLibCode = mController.Sample();

		// Check result of sample instruction
		if (wiiLibCode == HB_SUCCESS)
		{
			// Check for updated button states
		    if (mController.HasButtonData()	&&
				mController.mStatus.buttons != mButtonStates)
		    {
				// Update button states
				mButtonStates = mController.mStatus.buttons;

				// Fire button states changed event
				mButtonStatesChangedEvent->Fire(this, NULL);
			}

			// Check for other updates if motion data is present
			if (mController.HasMotionData())
			{
				// Process sample in tilt class
				mEstTilt.ProcessSample(mController.mStatus.accNatural);

				// Get new orientation
				mEstTilt.GetAxisAngle(wiiLibAxis, wiiLibAngle);
				newOrientation = CORE_3DOT::quat(ConvWiiLibVec3(wiiLibAxis),
												 wiiLibAngle);

				// Check if orientation has changed
				if (newOrientation.GetS() != mOrientation.GetS() ||
					newOrientation.GetX() != mOrientation.GetX() ||
					newOrientation.GetY() != mOrientation.GetY() ||
					newOrientation.GetZ() != mOrientation.GetZ())
				{
					// Update orientation
					mOrientation = newOrientation;

					// Fire orientation changed event
					mOrientationChangedEvent->Fire(this, NULL);
				}

				// Process sample in position class
				mEstTilt.GetGravityEstimate(wiiLibGravity);
				mEstPosition.ProcessSample(mController.mStatus.accNatural,
										   wiiLibGravity, DELTA_T);

				// Get new position
				mEstPosition.GetPosition(wiiLibPosition);
				newPosition = ConvWiiLibVec3(wiiLibPosition);

				// Check if position has changed
				if (newPosition.GetX() != mPosition.GetX() ||
					newPosition.GetY() != mPosition.GetY() ||
					newPosition.GetZ() != mPosition.GetZ())
				{
					// Update position
					mPosition = newPosition;
	
					// Fire position changed event
					mPositionChangedEvent->Fire(this, NULL);
				}
			} // if (mController.HasMotionData())
		}
		else // if (wiiLibCode == HB_SUCCESS)
		{
			LOG("WiiControllerExtension") << "Controller::Sample() Errorcode: "
				<< wiiLibCode << LOGEND();
		}
	}

	/*
	 * Exposed methods
	 */

	// Returns orientation of controller (angle-axis => quaternion)
	CORE_3DOT::quat WiiControllerExtension::GetOrientation() const
	{
		return mOrientation;
	}

	// Returns position of controller
	CORE_3DOT::vec3 WiiControllerExtension::GetPosition() const
	{
		return mPosition;
	}

	/*
	 * Event handles
	 */

	// Returns button states changed event handle
	CORE_3DOT::HEvent WiiControllerExtension::GetButtonStatesChangedEvent()
	{
		return mButtonStatesChangedEvent;
	}

	// Returns orientation changed event handle
	CORE_3DOT::HEvent WiiControllerExtension::GetOrientationChangedEvent()
	{
		return mOrientationChangedEvent;
	}

	// Returns position changed event handle
	CORE_3DOT::HEvent WiiControllerExtension::GetPositionChangedEvent()
	{
		return mPositionChangedEvent;
	}
}
